#!/usr/local/bin/perl

# Sendmail "socket" map script to perform SRS functions.
#
# Code by Mark Kramer <admin@asarian-host.net>
#
# Version 0.32rc3
#
# Last revision: November 11, 2004
#
# With thanks to Jim Allen for pointing out a missing
# "while (!eof($client))" loop, causing BROKEN PIPE errors
# on concurrent connections.
#
# Licensed under GPL
#
# For detailed installation notes, read:
#
#           http://srs-socketmap.info/documentation.html
#           INSTALL notes in tarball
#
# See also: http://www.libsrs2.org/
#           http://spf.pobox.com/
#
# This version requires at least Sendmail 8.13.0 + Mail::SRS 0.30
#
# Revisions:
#	2012-04-02 jk@zid.tuwien.ac.at
#		Fixed: make_srs may receive blanks in address, retain quoting.
#		Fixed: exit read-in loop if some error condition arises.	
#

use IO::Socket;
use POSIX qw (:sys_wait_h);
use Sys::Syslog qw (:DEFAULT setlogsock);
use Mail::SRS;
use Getopt::Std;
use File::stat;
use strict;

my ($user, $login, $pass, $uid, $gid, $data, $socket_map, $old_address, $new_address, $use_address, $client, $sock, $config_file, $socket_path, $pid_path, $secret, $fwdomain, $max_age, $hash_length);
my $Version = 'srs-socketmapd V-0.32rc3';
my %options = ();
my %Config_Params = ();
my $config_file_main = '/etc/srs-socketmapd.conf';
my $config_file_dist = '/usr/local/etc/srs-socketmapd.conf';

# Get cmdline

getopts ( "VC:u:", \%options);

# If get version (-V) print version and exit

if ($options{V}) {
    print STDERR "$Version\n";
    exit 0;
}

# startup

init_syslog ();

write_syslog ("srs-socketmapd starting up");

parse_config ();

my $srs_all = new Mail::SRS (Secret => $secret, MaxAge => $max_age, HashLength => $hash_length, AlwaysRewrite => 1);
my $srs = new Mail::SRS (Secret => $secret, MaxAge => $max_age, HashLength => $hash_length);

# Subroutines

sub parse_config {

    # Alternate config

    if ($options{C}) {
        print STDERR "Alternate config file specified : $options{C}\n";
        $config_file = $options{C};
    }

    # Check for conf files

    if (not open (CONFIG_TEST, "< $config_file_main")) {
        $config_file_main ='';
    } else {
        close (CONFIG_TEST);
    }
    if (not open (CONFIG_TEST, "< $config_file_dist")) {
        $config_file_dist ='';
    } else {
        close (CONFIG_TEST);
    }

    # Select configuration whichever one found first in order

    if ($config_file) {
    } elsif ($config_file_main) {
        $config_file = $config_file_main;
    } elsif ($config_file_dist) {
        $config_file = $config_file_dist;
    }

    # Read config

    if ($config_file) {
        my $info = stat ($config_file) or die "Unable to stat $config_file";

        # Config file must be owned by root

        die "$config_file is not owned by root!" if ($info -> uid);

        # group/world must not have read/write access to it
        # Config file has our secret!

        die "Unsafe permissions on $config_file!" if ($info -> mode & 066);
        write_syslog ("Using configuration from <$config_file>");
        open (CONFIG, "< $config_file") or die "Error opening <$config_file> for reading";
        while (<CONFIG>) {
            chomp;
            s/#.*//;
            s/^\s+//;
            s/\s+$//;
            next unless length;
            my ($var, $value) = split (/\s*=\s*/, $_, 2);
            $Config_Params{uc ($var)} ||= $value;
        }
        close (CONFIG);
    } else {
        log_error_and_exit ("ERROR -- no config found");
    }

    # Set config and use defaults

    $user = $options{u} || $Config_Params{"USER"} ;  # cmdlind overrides conf
    $socket_path = $Config_Params{"SOCKETPATH"} || "/var/run/srs-socketmapd.sock";
    $pid_path = $Config_Params{"PIDPATH"} || "/var/run/srs-socketmapd.pid";
    $secret = $Config_Params{"SECRET"};
    $fwdomain = $Config_Params{"ALIAS"};
    $max_age = $Config_Params{"MAXAGE"} || 8;
    $hash_length = $Config_Params{"HASHLENGTH"} || 8;
    init_syslog($Config_Params{"LOGFACILITY"}) if $Config_Params{"LOGFACILITY"};
}

sub init_syslog {
    my $facility = shift;
    $facility = "mail" if ! defined $facility;

    closelog ();
    setlogsock ('unix');
    openlog ('srs-socketmapd', 'pid,cons', $facility) or exit 1;
}

sub write_syslog {
    syslog ('info', @_);
}

sub log_error_and_exit {
    write_syslog (@_);
    print STDERR "@_\n";
    exit 1;
}

sub netstringRead {
    my $sock = shift;
    my $saveSeparator = $/;
    $/ = ':';
    my $dataLength = <$sock>;
    write_syslog ("WARNING: cannot read netstring length") unless defined ($dataLength);
    chomp ($dataLength);
    my $data;
    if ($sock -> read ($data, $dataLength) == $dataLength) {
        ($sock -> getc () eq ',') or write_syslog ("WARNING: data misses closing ,");
    } else {
        write_syslog ("WARNING: received only " . length ($data) . " of $dataLength bytes");
    }
    $/ = $saveSeparator;
    return $data;
}

sub netstringWrite {
    my $sock = shift;
    my $data = "OK " . shift;
    write_syslog ("WARNING: $@") if (not eval {print $sock length ($data) . ':' . $data . ','});
}

sub handleChildConnection {
    my $client = shift;
    while (not eval {eof ($client)}) {
        if (eval {$data = netstringRead ($client)}) {
            if ($data =~ /^(\S+)\s(.+)$/) {
                $socket_map = $1;
                $old_address = $2;
               ($use_address = $old_address) =~ s/[<>]//g;
                $use_address =~ s/\.$//g;
                if ($socket_map eq 'reverse_srs') {
                    if ($use_address =~ /^SRS[01][-+=]/i) {
                        $use_address = $_ while (eval {$_ = $srs -> reverse ($use_address)});
                        $use_address .= '.>';
                        $use_address =~ s/\@/<@/;
                        $use_address =~ s/^([^<@]* .*?)<@/"$1"<@/;      # quote if blank inside!
                        netstringWrite ($client, $use_address);
                    } elsif ($use_address =~ /^\|/) {
                        netstringWrite ($client, "\"$old_address\"");
                    } else {
                        netstringWrite ($client, $old_address);
                    }
                } elsif ($socket_map eq 'make_srs') {
                    if (eval {$new_address = $srs -> reverse ($use_address)}) {
                        netstringWrite ($client, $old_address);
                    } elsif (eval {$new_address = $srs -> forward ($use_address, $fwdomain)}) {
                        $new_address .= '.>';
                        $new_address =~ s/\@/<@/;
                        $new_address =~ s/^([^<@]* .*?)<@/"$1"<@/;      # quote if blank inside!
                        netstringWrite ($client, $new_address);
                    } else {
                        netstringWrite ($client, $old_address);
                    }
                } elsif  ($socket_map eq 'make_srs_all') {
                    if (eval {$new_address = $srs_all -> reverse ($use_address)}) {
                        netstringWrite ($client, $old_address);
                    } elsif (eval {$new_address = $srs_all -> forward ($use_address, $fwdomain)}) {
                        $new_address .= '.>';
                        $new_address =~ s/\@/<@/;
                        $new_address =~ s/^([^<@]* .*?)<@/"$1"<@/;      # quote if blank inside!
                        netstringWrite ($client, $new_address);
                    } else {
                        netstringWrite ($client, $old_address);
                    }

                } else {
                    write_syslog ("WARNING: unknown socketmap, '$socket_map'");
                    return;
                }
            } else {
                write_syslog ("WARNING: incomplete data, '$data'");
                return;
            }
        } else {
            write_syslog ("WARNING: unable to read from client");
            return;
        }
    } # while
}

# So far so good, lets fire her up

# Check to see if we have enough to run

if ((not $user) || (not $secret) || (not $fwdomain)) {
    print STDERR "Missing user, secret or alias/fwdomain -- CHECK config file\n";
    print STDERR "Usage: $0 -u <user to run as> -C <config file>\n";
    exit 1;
} elsif ($>) {
    print STDERR "You need to start socketmapd as root!\n";
    exit 1;
} else {
   ($login, $pass, $uid, $gid) = getpwnam ($user);
    if (not defined ($uid)) {
        log_error_and_exit ("<$user> is not a valid user on this system!");
    } elsif (not $uid) {
        log_error_and_exit ("You cannot run socketmapd as root!");
    }
}

open (STDIN, '/dev/null');
open (STDOUT, '>/dev/null');
open (STDERR, '>&STDOUT');

umask (0027);

unlink ($pid_path);
unlink ($socket_path);

if ($_ = fork ()) {
    open (USERLOG, ">".$pid_path) or exit 1;
    flock (USERLOG, 2);
    seek (USERLOG, 0, 0);
    print USERLOG " $_";
    close (USERLOG);
    exit 0;
}

POSIX::setsid () || exit 1;

if (not (eval {$sock = new IO::Socket::UNIX (
              Listen => SOMAXCONN,
                Type => SOCK_STREAM,
               Local => $socket_path)})) {
    log_error_and_exit ("ERROR: Unable to create UNIX domain socket!");
}

chown $uid, $gid, $socket_path;

$0 = 'srs-socketmapd';

$) = $gid;
$( = $gid;
$> = $uid;
$< = $uid;

write_syslog ("Dropped privileges on srs socketmap daemon");

write_syslog ("srs-socketmapd running");

while (eval {$client = $sock -> accept ()}) {
    if (fork) {
        eval {$client -> close ()};
        wait;
    } elsif (fork) {
        exit 0;
    } else {
        eval {$sock -> close ()};
        eval {handleChildConnection ($client)};
        eval {$client -> close ()};
        exit 0;
    }
}

write_syslog ("Unsuccessful exit from the srs socketmap daemon: $!");

eval {$sock -> close ()};

exit 0;

